import numpy as np
import pandas as pd
from sklearn.ensemble import RandomForestRegressor
from sklearn.metrics import (
    mean_absolute_error, root_mean_squared_error,
    mean_squared_log_error, mean_absolute_percentage_error,
    max_error, r2_score, explained_variance_score
)





class RandomRiver:

    def __init__(
        self,
        df: pd.core.frame.DataFrame,
        date_col: str,
        y_col: str,
        x_var_list: list = None,
        split_frac: float = 0.8
    ) -> None:
        self.__df = df
        self.__date_col = date_col
        self.__y_col = y_col
        self.__x_var_list = x_var_list
        self.__split_frac = split_frac
        self.__model = None


    def __repr__(self) -> str:
        if self.__x_var_list:
            lst_str = "['"+ "', '".join(self.__x_var_list) + "']"
            return f"RandomRiver(date_col = '{self.__date_col}', y_col = '{self.__y_col}', x_var_list = {lst_str})"
        else:
            return f"RandomRiver(date_col = '{self.__date_col}', y_col = '{self.__y_col}')"


    def __tryel(func):
        def wrapper(self, *args, **kwargs):
            try:
                return func(self, *args, **kwargs)
            except Exception as e:
                print(f"Method '{func.__name__}' failed with Exception:\n{e}")
        return wrapper


    def __split_data(self) -> list:
        flow_df = self.flow
        train = flow_df.sample(frac = self.__split_frac)
        train_X = train.drop(self.__y_col, axis = 1).reset_index(drop = True)
        train_y = train[self.__y_col].reset_index(drop = True)
        test = flow_df.drop(train.index)
        test_X = test.drop(self.__y_col, axis = 1).reset_index(drop = True)
        test_y = test[self.__y_col].reset_index(drop = True)
        return [train_X, train_y, test_X, test_y]


    def __flow(self, df, test_data = False) -> pd.core.frame.DataFrame:
        date_df = pd.DataFrame({
            "year": pd.to_datetime(df[self.__date_col]).dt.year,
            "month": pd.to_datetime(df[self.__date_col]).dt.month,
            "day": pd.to_datetime(df[self.__date_col]).dt.day
        })
        if test_data:
            if self.__x_var_list:
                return pd.concat([date_df, df[self.__x_var_list]], axis = 1)
            else:
                return date_df
        else:
            if self.__x_var_list:
                return pd.concat([date_df, df[(self.__x_var_list + [self.__y_col])]], axis = 1)
            else:
                return pd.concat([date_df, df[self.__y_col]], axis = 1)


    @property
    def flow(self) -> pd.core.frame.DataFrame:
        date_df = pd.DataFrame({
            "year": pd.to_datetime(self.__df[self.__date_col]).dt.year,
            "month": pd.to_datetime(self.__df[self.__date_col]).dt.month,
            "day": pd.to_datetime(self.__df[self.__date_col]).dt.day
        })
        if self.__x_var_list:
            return pd.concat([date_df, self.__df[(self.__x_var_list + [self.__y_col])]], axis = 1)
        else:
            return pd.concat([date_df, self.__df[self.__y_col]], axis = 1)


    def fit(
        self,
        rf_n_estimators: int = 200,
        rf_criterion: str = "squared_error",
        rf_max_depth: int = None,
        rf_min_samples_split: int | float = 2
    ):
        train_X, train_y, _, _ = self.__split_data()
        self.__model = RandomForestRegressor(
            n_estimators = rf_n_estimators,
            criterion = rf_criterion,
            max_depth = rf_max_depth,
            min_samples_split = rf_min_samples_split
        )
        return self.__model.fit(train_X, train_y)


    def accuracy(self, return_accuracy_json: bool = False):
        if self.__model:
            _, _, test_X, test_y = self.__split_data()
            rf_pred = self.__model.predict(test_X)
            if return_accuracy_json:
                return {
                    "R2": r2_score(test_y, rf_pred),
                    "RMSE": root_mean_squared_error(test_y, rf_pred),
                    "MAE": mean_absolute_error(test_y, rf_pred),
                    "MSLE": mean_squared_log_error(test_y, rf_pred),
                    "MAPE": mean_absolute_percentage_error(test_y, rf_pred),
                    "MaxError": max_error(test_y, rf_pred),
                    "ExplainedVarianceScore": explained_variance_score(test_y, rf_pred)
                }
            else:
                return pd.DataFrame({
                    "R2": [r2_score(test_y, rf_pred)],
                    "RMSE": [root_mean_squared_error(test_y, rf_pred)],
                    "MAE": [mean_absolute_error(test_y, rf_pred)],
                    "MSLE": [mean_squared_log_error(test_y, rf_pred)],
                    "MAPE": [mean_absolute_percentage_error(test_y, rf_pred)],
                    "MaxError": [max_error(test_y, rf_pred)],
                    "ExplainedVarianceScore": [explained_variance_score(test_y, rf_pred)]
                })
        else:
            print("No model exists - use the '.fit()' method to train one")
            if return_accuracy_json:
                return {
                    "R2": pd.NA,
                    "RMSE": pd.NA,
                    "MAE": pd.NA,
                    "MSLE": pd.NA,
                    "MAPE": pd.NA,
                    "MaxError": pd.NA,
                    "ExplainedVarianceScore": pd.NA
                }
            else:
                return pd.DataFrame({
                    "R2": [pd.NA],
                    "RMSE": [pd.NA],
                    "MAE": [pd.NA],
                    "MSLE": [pd.NA],
                    "MAPE": [pd.NA],
                    "MaxError": [pd.NA],
                    "ExplainedVarianceScore": [pd.NA]
                })


    def predict(self, new_df):
        x_var_df = self.__flow(new_df, test_data = True)
        return self.__model.predict(x_var_df)



import os
import json
import time
import numpy as np
import pandas as pd
from ollama import Client
from pdfminer.high_level import extract_pages



class Ragby:

    def __init__(self, chat_model, embedding_model, top_n = 10, ollama_client = None, data_dir = None):
        self.chat_model = chat_model
        self.embedding_model = embedding_model
        self.top_n = top_n
        self.ollama_client = Client(host = "http://localhost:11434") if ollama_client is None else ollama_client
        self.data_dir = "data" if data_dir is None else data_dir
        self.__setup_dirs()


    def __repr__(self):
        return f"Ragby(chat_model={self.chat_model}, embedding_model={self.embedding_model}, top_n={self.top_n}, ollama_client={self.ollama_client})"


    def __setup_dirs(self):
        if "chunks" not in os.listdir():
            print("--- Creating 'chunks' folder")
            os.mkdir("chunks")
        if "embeddings" not in os.listdir():
            print("--- Creating 'embeddings' folder")
            os.mkdir("embeddings")

    
    def make_chunks(self, file_name, paragraphs = False):
        file_name_prefix = file_name.split(".")[0]
        data_file_path = f"{self.data_dir}/{file_name}"
        if paragraphs:
            with open(data_file_path, "r") as f:
                chunks = []
                buffer = []
                for line in f.readlines():
                    line = line.strip()
                    if line:
                        buffer.append(line)
                    elif len(buffer):
                        chunks.append(" ".join(buffer))
                        buffer = []
                if len(buffer):
                    chunks.append(" ".join(buffer))
        else:
            with open(data_file_path, "r") as f:
                chunks = f.read().splitlines()
        with open(f"chunks/{file_name_prefix}.json", "w") as f:
            json.dump(chunks, f)
            print(f"--- Created chunks of length: {len(chunks)}")
            print(f"--- Chunks stored in 'chunks/{file_name_prefix}.json'")
            return [chunks, f"chunks/{file_name_prefix}.json"]


    def make_chunks_pdf(self, file_name):
        file_name_prefix = file_name.split(".")[0]
        data_file_path = f"{self.data_dir}/{file_name}"
        pdf_elements = []
        for page_layout in extract_pages(data_file_path):
            for element in page_layout:
                pdf_elements.append(element)
        text_pdf_elements = [e for e in pdf_elements if str(type(e)) == "<class 'pdfminer.layout.LTTextBoxHorizontal'>"]
        text_pdf_elements = [e.get_text().strip() for e in text_pdf_elements]
        with open(f"chunks/{file_name_prefix}.json", "w") as f:
            json.dump(text_pdf_elements, f)
            print(f"--- Created chunks of length: {len(text_pdf_elements)}")
            print(f"--- Chunks stored in 'chunks/{file_name_prefix}.json'")
            return [text_pdf_elements, f"chunks/{file_name_prefix}.json"]

    
    def make_embeddings_return(self, c):
        if isinstance(c, list):
            return [self.ollama_client.embeddings(model = self.embedding_model, prompt = chunk)["embedding"] for chunk in c]
        elif isinstance(c, str):
            return self.ollama_client.embeddings(model = self.embedding_model, prompt = c)["embedding"]
        else:
            raise ValueError("Input to 'c' must be a string (str) or a list of strings (list[str])")

    
    def make_embeddings(self, c, file_name):
        file_name_prefix = file_name.split(".")[0]
        emb = [self.ollama_client.embeddings(model = self.embedding_model, prompt = chunk)["embedding"] for chunk in c]
        with open(f"embeddings/{file_name_prefix}.json", "w") as f:
            json.dump(emb, f)
            print(f"--- Created embeddings of length: {len(emb)}")
            print(f"--- Embeddings stored in 'embeddings/{file_name_prefix}.json'")
            return f"embeddings/{file_name_prefix}.json"
    
    
    def cosine_similarity(self, x, y_arr):
        x_norm = np.linalg.norm(x)
        return [np.dot(x, y) / (x_norm * np.linalg.norm(y)) for y in y_arr]
    
    
    def get_embedding_similarity(self, sim_score_list, text_chunks):
        sorted_sim_scores = sorted(zip(sim_score_list, range(len(sim_score_list))), reverse = True)
        return [text_chunks[sss[1]] for sss in sorted_sim_scores[:self.top_n]]
    
    
    def make_chat_message(self, user_prompt, system_prompt, text_chunks_path, text_embeddings_path):
        with open(text_embeddings_path, "r") as f:
            text_embeddings = json.load(f)
        with open(text_chunks_path, "r") as f:
            text_chunks = json.load(f)
        prompt_embedding = self.ollama_client.embeddings(model = self.embedding_model, prompt = user_prompt)["embedding"]
        sim_scores = self.cosine_similarity(prompt_embedding, text_embeddings)
        similar_chunks = self.get_embedding_similarity(sim_scores, text_chunks)
        system_content = system_prompt + "\n".join(similar_chunks)
        return [
            {
                "role": "system",
                "content": system_content
            },
            {
                "role": "user",
                "content": user_prompt
            }
        ]
    
    
    def chat(self, user_prompt, system_prompt, text_chunks_path, text_embeddings_path):
        chat_messages = self.make_chat_message(
            user_prompt = user_prompt, 
            system_prompt = system_prompt, 
            text_chunks_path = text_chunks_path, 
            text_embeddings_path = text_embeddings_path
        )
        response = self.ollama_client.chat(
            model = self.chat_model,
            messages = chat_messages
        )
        return response["message"]["content"]
